import 'dart:io';

import 'package:camera/camera.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:image/image.dart' as img;
import 'package:flutter/material.dart';
import 'package:google_mlkit_text_recognition/google_mlkit_text_recognition.dart';

import 'package:path_provider/path_provider.dart';
import 'package:reader_ocr/enum/reader_type.dart';
import 'package:reader_ocr/widgets/painter/interest_zone.dart';

import 'widgets/camera_viewer.dart';

class OcrViewer extends StatelessWidget {
  final TextRecognizer textRecognizer =
      TextRecognizer(script: TextRecognitionScript.latin);
  final ReaderType readerType;
  final double heightInterestFactor;
  final double widthInterestFactor;
  final int limitOfMatches;
  final Color ocrBorderColor;
  final Color ocrTextColor;
  final Function(String? value) onValueReaded;
  final Widget? child;
  OcrViewer(
      {super.key,
      required this.readerType,
      required this.onValueReaded,
      this.child,
      this.limitOfMatches = 3,
      this.heightInterestFactor = 0.2,
      this.widthInterestFactor = 0.7,
      this.ocrTextColor = Colors.blue,
      this.ocrBorderColor = Colors.blue});

  final ValueNotifier<String?> value = ValueNotifier<String?>(null);
  final ValueNotifier<bool> isProcessing = ValueNotifier<bool>(false);

  static img.Image convertYUV420ToImage(CameraImage cameraImage) {
    final imageWidth = cameraImage.width;
    final imageHeight = cameraImage.height;

    final yBuffer = cameraImage.planes[0].bytes;
    final uBuffer = cameraImage.planes[1].bytes;
    final vBuffer = cameraImage.planes[2].bytes;

    final int yRowStride = cameraImage.planes[0].bytesPerRow;
    final int yPixelStride = cameraImage.planes[0].bytesPerPixel!;

    final int uvRowStride = cameraImage.planes[1].bytesPerRow;
    final int uvPixelStride = cameraImage.planes[1].bytesPerPixel!;

    final image = img.Image(width: imageWidth, height: imageHeight);

    for (int h = 0; h < imageHeight; h++) {
      int uvh = (h / 2).floor();

      for (int w = 0; w < imageWidth; w++) {
        int uvw = (w / 2).floor();

        final yIndex = (h * yRowStride) + (w * yPixelStride);

        // Y plane should have positive values belonging to [0...255]
        final int y = yBuffer[yIndex];

        // U/V Values are subsampled i.e. each pixel in U/V chanel in a
        // YUV_420 image act as chroma value for 4 neighbouring pixels
        final int uvIndex = (uvh * uvRowStride) + (uvw * uvPixelStride);

        // U/V values ideally fall under [-0.5, 0.5] range. To fit them into
        // [0, 255] range they are scaled up and centered to 128.
        // Operation below brings U/V values to [-128, 127].
        final int u = uBuffer[uvIndex];
        final int v = vBuffer[uvIndex];

        // Compute RGB values per formula above.
        int r = (y + v * 1436 / 1024 - 179).round();
        int g = (y - u * 46549 / 131072 + 44 - v * 93604 / 131072 + 91).round();
        int b = (y + u * 1814 / 1024 - 227).round();

        r = r.clamp(0, 255);
        g = g.clamp(0, 255);
        b = b.clamp(0, 255);

        image.setPixelRgb(w, h, r, g, b);
      }
    }

    return image;
  }

  static Future<InputImage> cropImage(
      (
        CameraImage image,
        RootIsolateToken token,
        double heightInterestFactor,
        double widthInterestFactor
      ) args) async {
    BackgroundIsolateBinaryMessenger.ensureInitialized(args.$2);
    img.Image convertedImage =
        img.copyRotate(convertYUV420ToImage(args.$1), angle: 90);

    Directory directory = await getApplicationDocumentsDirectory();

    File file =
        await File("${directory.path}/arquivo2.jpg").create(recursive: true);
    file.writeAsBytesSync(img.encodeJpg(img.copyCrop(
      convertedImage,
      x: (args.$1.height - (args.$1.height * args.$4)) ~/ 2,
      y: (args.$1.width - (args.$1.width * args.$3)) ~/ 2,
      height: (args.$1.width * args.$3).toInt(),
      width: (args.$1.height * args.$4).toInt(),
    )));

    return InputImage.fromFile(file);
  }

  @override
  Widget build(BuildContext context) {
    void setValue(String text) {
      value.value = text;
    }

    RegExp selectRegex() {
      switch (readerType) {
        case ReaderType.word:
          return RegExp(r'[A-Za-z]{3,}');
        case ReaderType.numeric:
          return RegExp(r'[0-9]+(\.[0-9]+)?');
        case ReaderType.phone:
          return RegExp(r'\b\d{3}[-.\s]?\d{3}[-.\s]?\d{4}\b');
        case ReaderType.email:
          return RegExp(r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Za-z]{2,}\b',
              multiLine: true);
      }
    }

    String filterRecognizedText(String recognizedText) {
      RegExp regexToAply = selectRegex();

      String filteredText = readerType == ReaderType.numeric
          ? recognizedText.replaceAll(",", ".")
          : recognizedText.split(" ").first;

      List<Match> matches = regexToAply.allMatches(filteredText).toList();

      String stringReturn = "";
      for (var match in matches) {
        stringReturn += "${match.group(0)} \n";

        if (matches.indexOf(match) == (limitOfMatches)) {
          return stringReturn;
        }
      }

      return stringReturn;
    }

    Future<void> analyseImage(CameraImage inputImage) async {
      isProcessing.value = true;

      RecognizedText textRecognized =
          await textRecognizer.processImage(await compute(cropImage, (
        inputImage,
        RootIsolateToken.instance!,
        heightInterestFactor,
        widthInterestFactor
      )));

      setValue(filterRecognizedText(textRecognized.text));

      onValueReaded(value.value);

      isProcessing.value = false;
    }

    void onImageStreamed(CameraImage cameraImage) {
      if (isProcessing.value == false) {
        analyseImage(cameraImage);
      }
    }

    return Scaffold(
      body: CameraViewer(
        onImageStreamed: onImageStreamed,
        widget: Stack(
          children: [
            Center(
                child: CustomPaint(
              painter: InterestZone(
                  heightInterestFactor: heightInterestFactor,
                  widthInterestFactor: widthInterestFactor,
                  borderColors: ocrBorderColor),
              child: ListenableBuilder(
                listenable: value,
                builder: (context, child) => Visibility(
                    visible: value.value != null,
                    child: Center(
                      child: Text(value.value ?? "",
                          style: TextStyle(color: ocrTextColor)),
                    )),
              ),
            )),
            Builder(builder: (context) {
              if (child == null) {
                return Container();
              }

              return child!;
            })
          ],
        ),
      ),
    );
  }
}
